#!/bin/bash

debug=0
slot_base=10
write_state=0
state_dir=diag

while getopts "db:rws:" o
do
    case $o in
	d) debug=1;;
	b) slot_base=$OPTARG;;
	r) read_state=1;;
	w) write_state=1;;
	s) state_dir=$OPTARG;;
    esac
done

shopt -s extglob

if [[ $slot_base != +([0-9]) ]]
then
    echo "error: -b argument must be a digit string"
    exit 1
fi

if [[ -n $read_state ]]
then
    if [[ -z $state_dir ]]
    then
	echo "error: if -r is set then -s must be set, too"
	exit 1
    elif [[ ! -d $state_dir ]]
    then
	echo "error: the state directory named by -s does not exist"
	exit 1
    elif [[ ! -e $state_dir/front-end.state.awk ]]
    then
	echo "error: front end state file does not exist in state directory"
	exit 1
    fi
fi

preamble=$(
    if [[ -n $read_state ]]
    then
	cat $state_dir/front-end.state.awk
    else
	echo "next_slot_loc = $slot_base;"
    fi
)

awk \
    -v slot_base=$slot_base \
    -v write_state=$write_state \
    -v state_dir=$state_dir \
"

BEGIN {
    $preamble
    debug = $debug;
}

"'function slot_alloc(ref_type) {
    new_slot = -1;

    for (new_slot in slot_free) {
        break;
    }

    if (new_slot == -1) {
        new_slot = next_slot_loc;
        ++ next_slot_loc;
    } else {
        delete slot_free[new_slot];
    }

    if (ref_type ~ /temp/) {
        slot_temp[new_slot] = 1;
        slot_status[new_slot] = "temp";
    }

    if (ref_type ~ /var/) {
        slot_var[new_slot] = 1;
        slot_status[new_slot] = "var";
    }

    return new_slot;
}

function dump_state() {
    system("mkdir -p " state_dir);
    state_file = state_dir "/front-end.state.awk";
    printf("next_slot_loc=%s\n", next_slot_loc) > state_file;
    for (s in slot_free) printf("slot_free[%s]=1\n", s) >> state_file;
    for (s in slot_temp) printf("slot_temp[%s]=1\n", s) >> state_file;
    for (s in slot_var ) printf("slot_var[%s]=1\n", s) >> state_file;
    for (s in slot_status) printf("slot_status[%s]=\"%s\"\n", s, slot_status[s]) >> state_file;
    for (s in slot_var_name) printf("slot_var_name[%s]=\"%s\"\n", s, slot_var_name[s]) >> state_file;
    for (n in var_slot) printf("var_slot[\"%s\"]=%s\n", n, var_slot[n]) >> state_file;
    for (s in slot_type) printf("slot_type[%s]=\"%s\"\n", s, slot_type[s]) >> state_file;
}

function slot_map() {
    printf("INSTR:print \"next_slot_loc=\",%s,\"\\n\"\n", next_slot_loc);
    for (i=slot_base; i<next_slot_loc; ++i) {
        if (slot_status[i] == "var") {
            printf("INSTR:print \"slot[%s]=%s:%s\\n\"\n",
                   i, slot_status[i], slot_var_name[i]);
        } else {
            printf("INSTR:print \"slot[%s]=%s\\n\"\n", i, slot_status[i]);
        }
    }
}

function display_degree() {
    printf("INSTR:print \"degree = \", degree, \"\\n\"\n");
}

function slot_gc() {
    for (slot in slot_temp) {
        slot_status[slot] = "free";
        slot_free[slot] = 1;
        delete slot_type[slot];
    }

    delete slot_temp;
}

function expr_split(expression, expr_type) {
    if (debug) printf("INSTR:expr_split(\"%s\",\"%s\")\n", expression, expr_type);

    expr_len = length(expression);

    if (expr_type ~ /^E_MULDIV$/ ) repl_pos = match(expression, /@[0-9]+(\*|\/)@[0-9]+[^=]*$/);
    if (expr_type ~ /^E_MULDIV2$/) repl_pos = match(expression, /@[0-9]+(\*|\/)@[0-9]+/);
    if (expr_type ~ /^E_ADDSUB$/ ) repl_pos = match(expression, /(^|[\(+-=])@[0-9]+(\+|-)@[0-9]+([\)+-]|$)/);
    if (expr_type ~ /^E_EXT$/    ) repl_pos = match(expression, /\[@[0-9]+(,@[0-9]+)*\]/);
    if (expr_type ~ /^E_UCN$/    ) repl_pos = match(expression, /@[0-9]+\^[^=]*$/);
    if (expr_type ~ /^E_UC2$/    ) repl_pos = match(expression, /@[0-9]+\^/);
    if (expr_type ~ /^E_BCN$/    ) repl_pos = match(expression, /@[0-9]+\^@[0-9]+[^=]*$/);
    if (expr_type ~ /^E_BC2$/    ) repl_pos = match(expression, /@[0-9]+\^@[0-9]+/);
    if (expr_type ~ /^E_PAR$/    ) repl_pos = match(expression, /\(@[0-9]+\)/);
    if (expr_type ~ /^E_VAR$/    ) repl_pos = match(expression, /[a-z]+[0-9]*/);
    if (expr_type ~ /^E_NG1$/    ) repl_pos = match(expression, /^-@[0-9]+/);
    if (expr_type ~ /^E_NG2$/    ) repl_pos = match(expression, /[\+\-\*\/\(\[,=]-@[0-9]+/);
    if (expr_type ~ /^E_IN1$/    ) repl_pos = match(expression, /^[0-9]+/);
    if (expr_type ~ /^E_IN2$/    ) repl_pos = match(expression, /[^@0-9][0-9]+/);
    if (expr_type ~ /^E_XTR$/    ) repl_pos = match(expression, /@[0-9]+#@[0-9]+/);

    repl_pos -= 1;
    repl_len = RLENGTH;

    begin = 1;
    len = repl_pos;
    head = substr(expression, begin, len);

    begin += repl_pos;
    len = repl_len;
    replace = substr(expression, begin, len);

    begin += repl_len;
    len = expr_len - begin + 1;
    tail = substr(expression, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:repl = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);
}

function finish() {
    print "DONE:";
    fflush();
}

function postamble(slot_num) {
    print "EXPR:" (slot_num ? (head "@" slot_num tail) : "");
    finish();
}

function postamble_2(slot_num_1, op, slot_num_2) {
    print "EXPR:" head "@" slot_num_1 op "@" slot_num_2 tail;
    finish();
}

# QUIT:

$1 ~ /^quit$/ {
    printf("INSTR:quit\n");
    finish();
    next;
}

# SLOT_MAP:

$1 ~ /^slot$/ {
    slot_map();
    printf("INSTR:ignore = prompt()\n");
    finish();
    next;
}

# DUMP:

$1 ~ /^dump$/ {
    dump_state();
    printf("INSTR:ignore = prompt()\n");
    finish();
    next;
}

# DISPLAY DEGREE:

$1 ~ /^degree$/ {
    display_degree();
    printf("INSTR:ignore = prompt()\n");
    finish();
    next;
}

# INTERACTIVE:

$1 ~ /^interactive$/ {
    printf("INSTR:interactive = 1\n");
    printf("INSTR:ignore = prompt()\n");
    finish();
    next;
}

# PROMPT:

$1 ~ /^$/ {
    slot_gc();
    printf("INSTR:ignore = prompt()\n");
    finish();
    next;
}

# POLYNOMIAL:

$1 ~ /^P@[0-9]+$/ {
    slot = $1;
    sub(/P@/, "", slot);
    var_name = slot_var_name[slot];
    if (slot_type[slot] == "extn" && slot_status[slot] == "var") {
        printf("INSTR:ignore = poly_e(%s)\n", slot);
        printf("INSTR:define var() { print \"%s\" }\n", var_name);
        printf("INSTR:ignore = poly_disp_e()\n");
    } else if (slot_type[slot] != "extn") {
        printf("INSTR:ignore = error_message(19)\n");
    } else if (slot_status[slot] != "var") {
        printf("INSTR:ignore = error_message(20)\n");
    }
    postamble(0);
    next;
}

# NORM:

$1 ~ /^N@[0-9]+$/ {
    slot = $1;
    sub(/N@/, "", slot);
    var_name = slot_var_name[slot];
    if (slot_type[slot] == "extn") {
        printf("INSTR:ignore = norm_e(%s)\n", slot);
    } else {
        printf("INSTR:ignore = error_message(21)\n");
    }
    postamble(0);
    next;
}

# DISCRIMINANT:

$1 ~ /^D\(\)$/ {
    printf("INSTR:ignore = discriminant()\n");
    postamble(0);
    next;
}

# INITIALIZE MULT ARRAY

$1 ~ /^@[0-9]+\*@[0-9]+=@[0-9]+$/ {

    split($1, part, /\*/);
    sub(/^@/, "", part[1]);
    l_arg = part[1];

    split(part[2], npart, /=/);
    sub(/^@/, "", npart[1]);
    m_arg = npart[1];
    sub(/^@/, "", npart[2]);
    r_arg = npart[2];

    if (slot_type[l_arg] != "extn") printf("INSTR:ignore = error_message(10)\n");
    if (slot_type[m_arg] != "extn") printf("INSTR:ignore = error_message(11)\n");
    if (slot_type[r_arg] != "extn") printf("INSTR:ignore = error_message(12)\n");

    printf("INSTR:ignore = initialize_mult_eee(%s,%s,%s)\n", l_arg, m_arg, r_arg);

    slot_gc();
    finish();
    next;
}

# INITIALIZE CONJ ARRAY

$1 ~ /^@[0-9]+\^@[0-9]+=@[0-9]+$/ {

    split($1, part, /\^/);
    sub(/^@/, "", part[1]);
    l_arg = part[1];

    split(part[2], npart, /=/);
    sub(/^@/, "", npart[1]);
    m_arg = npart[1];
    sub(/^@/, "", npart[2]);
    r_arg = npart[2];

    if (slot_type[l_arg] != "extn") printf("INSTR:ignore = error_message(16)\n");
    if (slot_type[m_arg] != "base") printf("INSTR:ignore = error_message(17)\n");
    if (slot_type[r_arg] != "extn") printf("INSTR:ignore = error_message(18)\n");

    printf("INSTR:ignore = initialize_conj_ebe(%s,%s,%s)\n", l_arg, m_arg, r_arg);

    slot_gc();
    finish();
    next;
}

# DISPLAY:

$1 ~ /^@[0-9]+$/ {
    slot = $1;
    sub(/^@/, "", slot);

    if (slot_type[slot] ~ /base/) printf("INSTR:ignore = display_b(%s)\n", slot);
    if (slot_type[slot] ~ /extn/) printf("INSTR:ignore = display_e(%s)\n", slot);

    postamble(0);
    next;
}

# PARENTHESIS:

$1 ~ /\(@[0-9]+\)/ {

    expr_split($1, "E_PAR");
    slot = replace;

    sub(/^\(@/, "", slot);
    sub(/\)$/,  "", slot);

    postamble(slot);
    next;
}

# VARIABLE:

$1 ~ /[a-z]+[0-9]*/ {

    expr_split($1, "E_VAR");
    var_name = replace;

    if (var_name in var_slot) {
        slot = var_slot[var_name];
    } else {
        slot = slot_alloc("var");
        var_slot[var_name] = slot;
        slot_var_name[slot] = var_name;
    }

    postamble(slot);
    next;
}

# INITIALIZE:

# The block below picks up literal values made up of digit strings
# which do *not* begin with a @.  This block handles digit strings
# that begin at the beginning of the expression (this is the first
# case) as well as digit strings that begin somewhere inside the
# expression (this is the second case).


$1 ~ /(^|[^@0-9])[0-9]+/ {
    expr_split($1, "E_IN1");

    if (replace ~ /^$/) {
       expr_split($1, "E_IN2");
       head = head substr(replace, 1, 1);
       sub(/^./, "", replace);
    }

    literal = replace;

    slot = slot_alloc("temp");
    slot_type[slot] = "base";
    printf("INSTR:ignore = initialize_bl(%s,%s)\n", slot, literal);

    postamble(slot);
    next;
}

# EXTENSION LITERAL:

$1 ~ /\[@[0-9]+(,@[0-9]+)*\]/ {

    expr_split($1, "E_EXT");

    slot = slot_alloc("temp");
    slot_type[slot] = "extn";

    printf("INSTR:ignore = reset_e(%s)\n", slot);
    sub(/\[/, "", replace);
    sub(/\]/, "", replace);
    n = split(replace, part, /,/);
    for (i=1; i<=n; ++i) {
        sub(/^@/, "", part[i]);
        if (slot_type[part[i]] == "base")
            printf("INSTR:ignore = initialize_ebl(%s,%s,%s)\n", slot, part[i], i);
        else {
            printf("INSTR:ignore = error_message(4)\n");
	    slot = 0;
        }
    }

    postamble(slot);
    next;
}

# COPY:

$1 ~ /^@[0-9]+=@[0-9]+$/ {
    if (debug) printf("INSTR:copy\n");

    split($1, part, /=/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);
    lvar = part[1];
    rvar = part[2];

    if (debug == 1) printf("INSTR:lvar = <%s>\n", lvar);
    if (debug == 1) printf("INSTR:rvar = <%s>\n", rvar);

    if (slot_status[lvar] == "var") {
        if (slot_type[rvar] == "base") { f = "copy_bb"; }
        if (slot_type[rvar] == "extn") { f = "copy_ee"; }
        printf("INSTR:ignore = %s(%s,%s)\n", f, lvar, rvar);
	slot_type[lvar] = slot_type[rvar];
    } else {
        printf("INSTR:print \"error: left operand is not a variable\\n\"\n");
    }

    postamble(0);
    next;
}

# BINARY CONJUGATE
$1 ~ /@[0-9]+\^@[0-9]+[^=]*$/ {

    expr_split($1, "E_BCN");
    con_head = head
    expr_split(replace, "E_BC2");
    head = con_head

    split(replace, part, /\^/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    slot = slot_alloc("temp");
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
        printf("INSTR:ignore = conjugate_eeb(%s,%s,%s)\n", slot, l_arg, r_arg);
        slot_type[slot] = "extn";
    } else if (slot_type[l_arg] == "base") {
        printf("INSTR:ignore = error_message(2)\n");
	slot = 0;
    } else {
        printf("INSTR:ignore = error_message(3)\n");
	slot = 0;
    }

    postamble(slot);
    next;
}

# UNARY CONJUGATE
$1 ~ /@[0-9]+\^[^=]*$/ {

    expr_split($1, "E_UCN");
    con_head = head;
    expr_split(replace, "E_UC2");
    head = con_head;

    sub(/^@/, "", replace);
    sub(/\^$/, "", replace);

    arg = replace;

    slot = slot_alloc("temp");
    if (slot_type[arg] == "extn") {
        printf("INSTR:ignore = initialize_bl(%s,1)\n", slot);
        slot_type[slot] = "base";
        postamble_2(arg, "^", slot);
    } else {
        printf("INSTR:ignore = error_message(1)\n");
        postamble(0);
    }

    next;
}

# NEGATE:

$1 ~ /(^|[\+\-\*\/\(\[,=])-@[0-9]+/ {

    expr_split($1, "E_NG1");

    if (replace ~ /^$/) {
       expr_split($1, "E_NG2");
       head = head substr(replace, 1, 1);
       sub(/^./, "", replace);
    }

    rvar = replace;
    sub(/^-@/, "", rvar);

    if (slot_type[rvar] == "base") { f = "negate_bb"; }
    if (slot_type[rvar] == "extn") { f = "negate_ee"; }
    slot = slot_alloc("temp");
    slot_type[slot] = slot_type[rvar];
    printf("INSTR:ignore = %s(%s,%s)\n", f, slot, rvar);

    postamble(slot);
    next;
}

# MULDIV

$1 ~ /@[0-9]+(\*|\/)@[0-9]+[^=]*$/ {

    expr_split($1, "E_MULDIV");
    mul_head = head;
    expr_split(replace, "E_MULDIV2");
    head = mul_head

    split(replace, part, /(\*|\/)/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    if (match(replace,/\*/)) {
	if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
	    f = "multiply_bbb";
	    derived_slot_type = "base";
	} else {
	    f = "multiply_eee";
	    derived_slot_type = "extn";
	}

	if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
	    temp_slot = slot_alloc("temp");
	    printf("INSTR:ignore = convert_eb(%s,%s)\n", temp_slot, l_arg);
	    l_arg = temp_slot;
	}

	if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
	    temp_slot = slot_alloc("temp");
	    printf("INSTR:ignore = convert_eb(%s,%s)\n", temp_slot, r_arg);
	    r_arg = temp_slot;
	}

	slot = slot_alloc("temp");
	printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
	slot_type[slot] = derived_slot_type;

	postamble(slot);
    } else {
	if (slot_type[r_arg] == "base") { f = "invert_bb"; }
	if (slot_type[r_arg] == "extn") { f = "invert_ee"; }
	slot = slot_alloc("temp");
	slot_type[slot] = slot_type[r_arg];
	printf("INSTR:ignore = %s(%s,%s)\n", f, slot, r_arg);

	postamble_2(l_arg, "*", slot);
    }

    next;
}

# ADDSUB:

$1 ~ /(^|[\(+-=])@[0-9]+(\+|-)@[0-9]+([\)+-]|$)/ {

    expr_split($1, "E_ADDSUB");

    if (match(replace, /^[\(+-=]/)) {
        head = head substr(replace,1,1);
        replace = substr(replace,2,length(replace)-1);
    }

    if (match(replace, /[\)+-]$/)) {
        tail = substr(replace,length(replace),1) tail;
        replace = substr(replace,1,length(replace)-1);
    }

    operation = (match(replace,/\+/)) ? "add" : "subtract";

    split(replace, part, /(\+|-)/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
	f = operation "_bbb";
	derived_slot_type = "base";
    } else {
	f = operation "_eee";
	derived_slot_type = "extn";
    }

    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
	temp_slot = slot_alloc("temp");
	printf("INSTR:ignore = convert_eb(%s,%s)\n", temp_slot, l_arg);
	l_arg = temp_slot;
    }

    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
	temp_slot = slot_alloc("temp");
	printf("INSTR:ignore = convert_eb(%s,%s)\n", temp_slot, r_arg);
	r_arg = temp_slot;
    }

    slot = slot_alloc("temp");
    printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
    slot_type[slot] = derived_slot_type;

    postamble(slot);
    next;
}

# EXTRACT
$1 ~ /@[0-9]+#@[0-9]+/ {

    expr_split($1, "E_XTR");

    split(replace, part, /#/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    slot = slot_alloc("temp");
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
        printf("INSTR:ignore = extract_beb(%s,%s,%s)\n", slot, l_arg, r_arg);
        slot_type[slot] = "base";
    } else if (slot_type[l_arg] == "base") {
        printf("INSTR:ignore = error_message(5)\n");
	slot = 0;
    } else {
        printf("INSTR:ignore = error_message(6)\n");
	slot = 0;
    }

    postamble(slot);
    next;
}

# CATCH UNMATCHED PATTERNS:

{
    printf("INSTR:ignore = error_message(0)\n");
    postamble(0);
    next;
}

END {
    if (write_state == 1) dump_state();
}

'
